----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local classHelper  = require("scenario/intersection/classHelper")
local classTimer   = require("scenario/intersection/classes/classTimer")

----------------------------------------------------------------------------------------------------

local ClassShockwave = classHelper.inheritsFrom(nil)

----------------------------------------------------------------------------------------------------
--Instance Methods
----------------------------------------------------------------------------------------------------

function ClassShockwave:hasFinished()
    return self.phase == "finished"
end

function ClassShockwave:resetShockwave()
    --[[
    Resets the shockwave
    ]]--
    for _, obj in ipairs(self.waveObjs) do
        obj.scale = Point3F(0, 0, 0);
    end
end

function ClassShockwave:reset()
    --[[
    Resets back to the initial state
    ]]--
    self.phase = "init"
    self:resetShockwave()
    self.timer:reset()
end

function ClassShockwave:begin(newPos)
    --[[
    Begins the shockwave animation
    Parameters:
        newPos - the new position
    ]]--
    errorHandler.assertNil(newPos)

    self.phase = "running"
    self.timer:setActive(true)

    self.position = newPos
end

function ClassShockwave:updateShockwave(baseScale, alpha)
    --[[
    Updates the shockwaves in-world
    Parameters:
        baseScale - the scale of the outer shockwave
        alpha     - the alpha of the colour of the shockwaves
    ]]--
    errorHandler.assertNil(baseScale, alpha)

    local position = self.position

    for i = 1, self.NUM_WAVES do
        local name = self.name..i

        local scale = baseScale/self.NUM_WAVES * i
        local colour = self.colourChange(i, alpha)

        --Set scale and alpha
        TorqueScript.eval(name..'.scale = "'..scale..' '..scale..' '..(i/self.NUM_WAVES)..'";')
        TorqueScript.eval(name..'.instanceColor = "'..colour..'";')
        TorqueScript.eval(name..'.position = "'..position:getX()..' '..position:getY()..' '..position:getZ()..'";')
    end
end

function ClassShockwave:update(dt)
    --[[
    Updates the shockwave in world
    Parameters:
        dt - the time since the last wave
    ]]--
    local timer = self.timer

    if self.phase == "running" then
        timer:update(dt)
        local time = timer:getTimePassed()

        local baseScale = self.RADIUS * (time/self.PERIOD)
        local alpha = math.max(0, (self.PERIOD - time) * self.MAX_ALPHA)

        self:updateShockwave(baseScale, alpha)
    end
end

----------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassShockwave:new(name, numWaves, radius, period, colourChange, callback)
    --[[
    Attributes:
        name         - the name of the shockwave
        numWaves     - the number of waves in the shockwave
        radius       - the maxium radius the shcokwave will grow to
        period       - the period of the shockwave animation
        colourChange - the function to calculate the shcokwaves colour change through time
                        - this function should take in:
                           - time
                           - alpha
                        - and return a string representing rgba colour
        callback     - the function to call on shockwave finish (can be nil)
    ]]--
    errorHandler.assertNil(name, numWaves, radius, period, colourChange)

    local self = {}
    self.position = nil
    self.name = name
    self.RADIUS = radius
    self.NUM_WAVES = numWaves
    self.PERIOD = period
    self.callback = callback

    self.MAX_ALPHA = 1
    self.colourChange = colourChange

    self.timer = classTimer:new(
        self.PERIOD,
        true,
        --Reset the shockwave and execute callback
        function()
            if self.callback then
                self.callback()
            end

            self.phase = "finished"
        end
    )

    self.phase = "init"

    self.waveObjs = {}
    for i = 1, self.NUM_WAVES do
        local waveName = self.name..i

        local obj = scenetree.findObject(waveName)
        --Create wave if it does not exist
        if not obj then
            TorqueScript.eval([[
              new TSStatic(]]..waveName..[[) {
                shapeName = "levels/smallgrid/art/intersection/pickups/explosives/cylinder.dae";
                dynamic = "1";
                scale = "0 0 0";
                position = "0 0 0";
                instanceColor = "0 0 0 0";
              };
            ]])

            obj = scenetree.findObject(waveName)
        end

        self.waveObjs[i] = obj
    end

    setmetatable(self, {__index = ClassShockwave})
    return self
end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassShockwave
