------------------------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local classHelper  = require("scenario/intersection/classHelper")

----------------------------------------------------------------------------------------------------

local ClassVector = classHelper.inheritsFrom(nil)

----------------------------------------------------------------------------------------------------
--Static Attributes
----------------------------------------------------------------------------------------------------

ClassVector.className = "ClassVector"

----------------------------------------------------------------------------------------------------
--Instance Methods
----------------------------------------------------------------------------------------------------

function ClassVector:getX() return self.components.x end
function ClassVector:getY() return self.components.y end
function ClassVector:getZ() return self.components.z end

function ClassVector:setX(newX)
  errorHandler.assertNil(newX)
  self.components.x = newX
end

function ClassVector:setY(newY)
  errorHandler.assertNil(newY)
  self.components.y = newY
end

function ClassVector:setZ(newZ)
  errorHandler.assertNil(newZ)
  self.components.z = newZ
end

function ClassVector:clone()
  return ClassVector:new(self.components.x, self.components.y, self.components.z)
end

function ClassVector:add(a)
  --[[
  Return the vector a added to this vector
  Parameters:
    a - the vector to add
  Returns:
    <ClassVector> - the resultant vector
  ]]--
  errorHandler.assertNil(a)

  if type(a) == "table" then
    --Resultant vector
    local r = ClassVector:new(0, 0, 0)
    r:setX(self.components.x + a:getX())
    r:setY(self.components.y + a:getY())
    r:setZ(self.components.z + a:getZ())
    return r
  else
    error("Data type error: "..type(a))
  end
end

function ClassVector:subtract(a)
  --[[
  Subtract vector a from this vector
  Parameters:
    a - the vector to subtract
  Returns:
    <ClassVector> - the resultant vector
  ]]--
  errorHandler.assertNil(a)

  if type(a) == "table" then
    local r = ClassVector:new(0, 0, 0)
    r:setX(self.components.x - a:getX())
    r:setY(self.components.y - a:getY())
    r:setZ(self.components.z - a:getZ())
    return r
  else
    error("Data type error: "..type(a))
  end
end

function ClassVector:multiply(a)
  --[[
  Multiply this vector by a vector/number a
  Parameters:
    a - the vector to use
  Returns:
    <ClassVector> - the resultant vector
  ]]--
  errorHandler.assertNil(a)

  local r = ClassVector:new(0, 0, 0)
  if type(a) == "table" then
    r:setX(self.components.x * a:getX())
    r:setY(self.components.y * a:getY())
    r:setZ(self.components.z * a:getZ())
  elseif type(a) == "number" then
    r:setX(self.components.x * a)
    r:setY(self.components.y * a)
    r:setZ(self.components.z * a)
  else
    error("Data type error: "..type(a))
  end

  return r
end

function ClassVector:divide(a)
  --[[
  Divide this vector by a vector/number a
  Parameters:
    a - the vector to divide
  Returns:
    <ClassVector> - the resultant vector
  ]]--
  errorHandler.assertNil(a)

  local r = ClassVector:new(0, 0, 0)
  if type(a) == "table" then
    r:setX(self.components.x / a:getX())
    r:setY(self.components.y / a:getY())
    r:setZ(self.components.z / a:getZ())
  elseif type(a) == "number" then
    r:setX(self.components.x / a)
    r:setY(self.components.y / a)
    r:setZ(self.components.z / a)
  else
    error("Data type error: "..type(a))
  end

  return r
end

function ClassVector:getMagnitude()
  --[[
  Returns:
    <number> - the length/magnitude of this vector
  ]]--
  return math.sqrt(math.pow(self.components.x, 2) + math.pow(self.components.y, 2) + math.pow(self.components.z, 2))
end

function ClassVector:getDistanceBetween(a)
  --[[
  Parameters:
    a - the vector to compare
  Returns:
    <number> - the distance between this vector and another
  ]]--
  errorHandler.assertNil(a)
  return self:subtract(a):getMagnitude()
end

function ClassVector:toUnitVector()
  --[[
  Returns:
    <ClassVector> - the unit vector of this vector ie. mag(v) = 1
  ]]--
  local mag = self:getMagnitude()
  return self:divide(mag)
end

function ClassVector:abs()
  --[[
  Absolutes each component and returns the resulting vector
  ]]--
  self.components.x = math.abs(self.components.x)
  self.components.y = math.abs(self.components.y)
  self.components.z = math.abs(self.components.z)
  return self
end

function ClassVector:isEqual(a)
  --[[
  Determine whether this vector is the same as vector a, or its magnitude is equal to a
  Parameters:
    a - the vector to compare
  Returns:
    <boolean> - if a is a vector, returns if all components match
                else            , returns if the magnitude match a
  ]]--
  errorHandler.assertNil(a)

  if type(a) == "table" then
    return self.components.x == a:getX() and self.components.y == a:getY() and self.components.z == a:getZ()
  else
    return self:getMagnitude() == a
  end
end

function ClassVector:isEqualTolerance(a, t)
  --[[
  Determine whether this vector is the same as vector a within a tolerance t
  Parameters:
    a - the vector to compare
    t - the tolerance
  Returns:
    <boolean> - if the components match within the specified tolerance
  ]]--
  errorHandler.assertNil(a, t)
  errorHandler.assertTrue(t >= 0, "Tolerance must be positive")

  if type(a) == "table" then
    return math.abs(self.components.x-a:getX()) <= t and
  math.abs(self.components.y-a:getY()) <= t and
  math.abs(self.components.z-a:getZ()) <= t
  else
    error("Data type error: "..type(a))
  end
end

function ClassVector:zero()
  --[[
  Sets all the vectors dimensions to zero
  ]]--
  self:setX(0)
  self:setY(0)
  self:setZ(0)
end

function ClassVector:removePlane(plane)
  --[[
  Converts a 3d vector to a 2d one by removing a plane
  Parameters:
    plane - the plane to remove
  Returns:
    <ClassVector> - the new 2d vector
  ]]--
  errorHandler.assertNil(plane)
  errorHandler.assertValidElement(plane, {"x", "y", "z"}, plane.." is an invalid plane")

  local a = self:clone()
  --Set the plane to 0
  a["set"..plane:upper()](a, 0)
  return a
end

function ClassVector:dot(u, v)
  --[[
  Calculates the dot product between 2 vectors
  Parameters:
    u - the first vector
    v - the other vector
  Returns:
    <number> - the dot product
  ]]--
  errorHandler.assertNil(u, v)

  return (u:getX() * v:getX()) + (u:getY() * v:getY()) + (u:getZ() * v:getZ())
end

function ClassVector:cross(u, v)
  --[[
  Calculates the cross product
  Parameters:
    u - the first vector
    v - the other vector
  Returns:
    <number> - the dot product
  ]]--
  errorHandler.assertNil(u, v)

  return ClassVector:new(
  u:getY()*v:getZ() - u:getZ()*v:getY(),
  u:getZ()*v:getX() - u:getX()*v:getZ(),
  u:getX()*v:getY() - u:getY()*v:getX()
  )
end

function ClassVector:getAngleBetween(a)
  --[[
  Calculates the angle between this vector and another
  Parameters:
    u - the first vector
    v - the other vector
  Returns:
    <number> - the angle
  ]]--
  errorHandler.assertNil(a)

  return math.acos(self:dot(self, a) / (self:getMagnitude() * a:getMagnitude()) )
end

function ClassVector:anglesBetween(a)
  --[[
  Calculates the xyz angles between this vector and another
  Parameters:
    a - the other vector
  Returns:
    angles - the table of euler angles
  ]]--
  errorHandler.assertNil(a)

  local angle = self:getAngleBetween(a)
  print(angle)
  local s, c = math.sin(angle), math.cos(angle)

  local axis = ClassVector:cross(self, a):toUnitVector()
  print(axis:toString())
  local x, y, z = axis:getX(), axis:getY(), axis:getZ()

  return {
  x = math.atan2(y * s - x * z * (1 - c) , 1 - (y^2 + z^2) * (1 - c)),
  y = math.asin (x * y * (1 - c) + z * s),
  z = math.atan2(x * s - y * z * (1 - c) , 1 - (x^2 + z^2) * (1 - c))
  }
end

function ClassVector:toTorquePos()
  --[[
  Returns:
    <string> - a string position vector or this vector
  ]]--
  return self.components.x..' '..self.components.y..' '..self.components.z
end

function ClassVector:toString()
  --[[
  Returns a string of this vector for viewing its attributes
  ]]--
  return '['..self.components.x..', '..self.components.y..', '..self.components.z..']'
end

------------------------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassVector:new(x, y, z)
  --[[
  Parameters:
    x - the x component
    y - the y component
    z - the z component
  ]]--
  errorHandler.assertNil(x, y, z)

  local self = {}
  setmetatable(self, {__index = ClassVector})
  self.components = {x = x, y = y, z = z}
  return self
end

------------------------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassVector
