------------------------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler              = require("scenario/intersection/handlers/errorHandler")
local ClassStandardCrashbreaker = require("scenario/intersection/classes/crashbreaker/classStandard")
local classHelper               = require("scenario/intersection/classHelper")
local scoringHandler            = require("scenario/intersection/handlers/scoringHandler")
local soundHandler              = require("scenario/intersection/handlers/soundHandler")

----------------------------------------------------------------------------------------------------

local ClassArmageddonCrashbreaker = classHelper.inheritsFrom(ClassStandardCrashbreaker)

----------------------------------------------------------------------------------------------------

function ClassArmageddonCrashbreaker:initEffects()
    --Create energy spheres
    for i = 1, 3 do
        local name = "crashbreaker_sphere_"..i
        if not scenetree.findObject(name) then
            TorqueScript.eval([[
              new TSStatic(]]..name..[[) {
                shapeName = "levels/smallgrid/art/intersection/pickups/explosives/sphere.dae";
                dynamic = "1";
                scale = "0 0 0";
                position = "0 0 0";
                instanceColor = "0 0 0 0";
              };
            ]])
        end
    end
end

function ClassArmageddonCrashbreaker:updateSpheres(scale, alpha)
    --[[
    Updates the energy spheres' position, scale and alpha
    Parameters:
        scale - the new scale (can be nil to stay the same)
        alpha - the new alpha (can be nil to stay the same)
    ]]--
    errorHandler.assertNil(scale, alpha)

    for i = 1, 3 do
        local name = "crashbreaker_sphere_"..i

        local col = self.constants.SPHERE_COLOURS[i]
        local colour = col.r.." "..col.g.." "..col.b.." "..alpha

        local pos = self.useData.position

        TorqueScript.eval(name..'.instanceColor = "'..colour..'";')
        TorqueScript.eval(name..'.scale = "'..scale..' '..scale..' '..scale..'";')
        TorqueScript.eval(name..'.position = "'..pos:getX()..' '..pos:getY()..' '..pos:getZ()..'";')
    end
end

function ClassArmageddonCrashbreaker:infuse()
    --[[
    Updates the shockwaves and the energy spheres
    ]]--
    local constants = self.constants
    local useData = self.useData

    local pos = useData.position

    --Update the shockwaves
    for i = 1, constants.NUM_SHOCKWAVES do
        local name = "crashbreaker_shockwave_"..i
        local alpha = math.min(1, math.max(0, constants.NUM_SHOCKWAVES*(useData.phaseTimer / constants.INFUSE_PERIOD) - i+1))

        if not scenetree.findObject(name) and alpha > 0 then
            local scale = constants.INFUSE_START_SCALE + i*0.2
            TorqueScript.eval([[
              new TSStatic(]]..name..[[) {
                shapeName = "levels/smallgrid/art/intersection/pickups/explosives/cylinder.dae";
                dynamic = "1";
                scale = "]]..scale..[[ ]]..scale..[[ ]]..(0.5 * i/10)..[[";
                position = "0 0 0";
                instanceColor = "0 0 0 0";
              };
            ]])
        end

        local colour = "1 "..(1 - i/constants.NUM_SHOCKWAVES).." 0 "..alpha

        TorqueScript.eval(name..'.position = "'..pos:getX()..' '..pos:getY()..' '..pos:getZ()..'";')
        TorqueScript.eval(name..'.instanceColor = "'..colour..'";')
    end


    --Shrink the energy spheres to the vehicle
    local ratio = (constants.INFUSE_PERIOD - useData.phaseTimer) / constants.INFUSE_PERIOD

    local scale = constants.STARTING_ENERGY_SPHERE_SCALE * ratio
    --An equation which starts a 0, curves to 1 and then back to 0 over the domain [0->1]
    local alpha = -(2*ratio - 1)^2 + 1

    self:updateSpheres(scale, alpha)
end

function ClassArmageddonCrashbreaker:explode()
    --[[
    Destroys the targets and animates the shockwave rings
    ]]--
    local data = self.useData

    local constants = self.constants
    local explosionPeriod = constants.EXPLOSION_PERIOD

    local ratio = data.phaseTimer/explosionPeriod.EXPLOSION * constants.EXPLOSION_RANGE

    --Update shockwaves
    for i = 1, constants.NUM_SHOCKWAVES do
        local name = "crashbreaker_shockwave_"..i
        local scale = (constants.INFUSE_START_SCALE + i*0.2) + ratio
        TorqueScript.eval(name..'.scale = "'..scale..' '..scale..' '..(0.5 * i/10)..'";')
    end

    --Affect vehicles within range
    for k, instance in pairs(self.vehicles) do
        if k ~= data.vehicleTrigger and not self.vehiclesAffected[k] then
            local dist = data.position:getDistanceBetween(instance:getPosition())

            if dist <= constants.EXPLOSION_RANGE * data.phaseTimer/explosionPeriod.TOTAL then
                instance:ignite()
                instance:breakAll()

                --Apply a force of max magnitude 1000 outwards for 1/4 of a second
                instance:boost(1000, data.position, 0.25)

                scoringHandler.addExplosion(instance)
                scoringHandler.addScore("collisions", self.constants.COLLISION_DAMAGE)

                instance:setCrashed()

                self.vehiclesAffected[k] = true
            end
        end
    end
end

function ClassArmageddonCrashbreaker:resetSpheres()
    --[[
    Deletes all shockwave objects
    ]]--
    for i = 1, 3 do
        local name = "crashbreaker_sphere_"..i
        if scenetree.findObject(name) then
            TorqueScript.eval(name..'.scale = "0 0 0";')
        end
    end
end

function ClassArmageddonCrashbreaker:resetEffects()
    --[[
    Resets all effects
    ]]--
    self:resetSpheres()
    self:resetShockwave()
end

-------------------------------------------------------------------------------------------------------
--Overridden Methods
--For method documentation, see superclass
----------------------------------------------------------------------------------------------------

function ClassArmageddonCrashbreaker:unlock()
    self.locked = false

    guihooks.trigger("i_messages_enqueue", {
        text = "ARMAGEDDON READY",
        cssClass = "extremeFlame",
        relPath = "pickups/armageddon.png",
        duration = 5,
        colour = "Red"
    })
end

function ClassArmageddonCrashbreaker:reset()
    self.useData.phase = "waiting"
    self.useData.phaseTimer = 0
    self:resetEffects()

    self.vehiclesCrashed = 0
    self.locked = true
    self.used = false
    self.inUse = false
end

function ClassArmageddonCrashbreaker:use(vehicle)
    if not self:canBeUsed() then
        return
    end

    errorHandler.assertNil(vehicle)

    self.inUse = true
    self:initEffects()

    --[[Set the explosive's position at the vehicle's
        and follow the vehicle's position for now]]--
    self.useData.position = vehicle:getPosition()
    self.useData.vehicleTrigger = vehicle:getId()

    guihooks.trigger("i_crashbreaker_update", {used = true})
    soundHandler.play("armageddon_ambient")

    self.useData.phase = "infuse"
end

function ClassArmageddonCrashbreaker:update(dt)
    local data = self.useData
    local constants = self.constants

    data.phaseTimer = data.phaseTimer + dt

    local phase = data.phase
    local timer = data.phaseTimer

    --Update applicable phase by calling its method and check to see whether it is time to move to the next phase
    if phase == "infuse" then
        self:infuse()

        if timer >= constants.INFUSE_PERIOD then
            data.phase = "explosion"
            data.phaseTimer = 0

            --[[ Pre-explosion sub-phase ]]--
            --Destroy activator vehicle
            local vehInst = self.vehicles[data.vehicleTrigger]
            vehInst:breakAll()
            --Break linked table
            data.position = data.position:clone()

            soundHandler.play("armageddon_explosion")
        end
    elseif phase == "explosion" then
        self:explode()

        --Final phase is up
        if timer >= constants.EXPLOSION_PERIOD.TOTAL then
            self.inUse = false
            self.used = true
            data.phaseTimer = 0
            data.phase = "finished"
            self:resetEffects()
        end
    end


    --Handle playing siren sounds when necessary
    local SIREN_INTERVALS = constants.SIREN_INTERVALS
    local NEXT_INT = SIREN_INTERVALS.INT[SIREN_INTERVALS.current]
    --PLay the next sound if passed interval bound
    if NEXT_INT and timer > NEXT_INT then
        SIREN_INTERVALS.current = SIREN_INTERVALS.current + 1
        soundHandler.play("armageddon_siren")
    end
end

function ClassArmageddonCrashbreaker:resetShockwave()
    for i = 1, self.constants.NUM_SHOCKWAVES do
        local name = "crashbreaker_shockwave_"..i
        if scenetree.findObject(name) then
            TorqueScript.eval(name..".delete();")
        end
    end
end

function ClassArmageddonCrashbreaker:instanceOf()
    return "ClassArmageddonCrashbreaker"
end

------------------------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassArmageddonCrashbreaker:new(vehicles, vehicleRequirement)
    --[[
    See ClassStandardCrashbreaker:new()
    ]]--
    local self = ClassStandardCrashbreaker:new(vehicles, vehicleRequirement)

    local constants = self.constants

    --How much score the explosion adds per vehicle
    constants.COLLISION_DAMAGE = 20000

    --The period of showing the shockwave of energy around the vehicle
    constants.INFUSE_PERIOD = 1.5
    constants.INFUSE_START_SCALE = 5
    --The period of the explosion (has sub-phases)
    constants.EXPLOSION_PERIOD = {
        IMPLOSION = 0.25,
        EXPLOSION = 0.25
    }
    constants.EXPLOSION_PERIOD.TOTAL = constants.EXPLOSION_PERIOD.IMPLOSION + constants.EXPLOSION_PERIOD.EXPLOSION
    constants.EXPLOSION_RANGE = 150

    constants.PERIOD = constants.INFUSE_PERIOD + constants.EXPLOSION_PERIOD.TOTAL

    --The starting scale of the energy spheres
    constants.STARTING_ENERGY_SPHERE_SCALE = 1000

    --The colours of the spheres at each stage (inner to outer)
    constants.SPHERE_COLOURS = {
        {r = 1, g = 1, b = 1},
        {r = 1, g = 1, b = 1},
        {r = 1, g = 1, b = 1}
    }

    --The number of shockwaves
    constants.NUM_SHOCKWAVES = 50

    --The intervals to play a siren sound
    constants.SIREN_INTERVALS = {INT = {0, 1}, current = 1}

    self.useData.phase = "waiting"
    --The time spent in the phase
    self.useData.phaseTimer = 0

    setmetatable(self, {__index = ClassArmageddonCrashbreaker})
    return self
end

------------------------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassArmageddonCrashbreaker