------------------------------------------------------------------------------------------------------------------
--Import
----------------------------------------------------------------------------------------------------

local classHelper       = require("scenario/intersection/classHelper")
local errorHandler      = require("scenario/intersection/handlers/errorHandler")
local soundHandler      = require("scenario/intersection/handlers/soundHandler")
local scoringHandler    = require("scenario/intersection/handlers/scoringHandler")

local ClassShockwave    = require("scenario/intersection/classes/classShockwave")
local ClassCrashbreaker = require("scenario/intersection/classes/crashbreaker/classCrashbreaker")

------------------------------------------------------------------------------------------------------------------

local ClassStandardCrashbreaker = classHelper.inheritsFrom(ClassCrashbreaker)

------------------------------------------------------------------------------------------------------------------
--Instance Methods
----------------------------------------------------------------------------------------------------

function ClassStandardCrashbreaker:update(dt)
    --[[
    Updates the explosion effect
    Parameters:
        dt - the time since last frame
    ]]--
    if self.locked then
        return
    end

    local data = self.useData
    local constants = self.constants

    data.timer = data.timer + dt

    --The scale of the outer cloud
    local baseScale = constants.EXPLOSION_RANGE * (data.timer/constants.PERIOD)
    --Fade out alpha as timer increases
    self.shockwave:updateShockwave(baseScale, math.max(0, (constants.PERIOD - data.timer) * constants.MAX_ALPHA))

    --Affect the vehicles
    for k, instance in pairs(self.vehicles) do
        --Do not affect the vehicle who triggered the explosion and those already under the effect
        if k ~= data.vehicleTrigger and not self.vehiclesAffected[k] then
            local distVec = instance:getPosition():subtract(data.position)
            local dist = distVec:getMagnitude()

            if dist <= baseScale/2 then
                --Apply a force of max magnitude 300 outwards for 1/4 of a second
                instance:boost(300 * (1-dist/constants.EXPLOSION_RANGE/2), data.position, 0.25)
                scoringHandler.addExplosion(instance)
                instance:setCrashed()

                self.vehiclesAffected[k] = true
            end
        end
    end

    --Timer is up
    if data.timer >= constants.PERIOD then
        self.inUse = false
        self.used = true
        self.shockwave:reset()
    end
end

function ClassStandardCrashbreaker:instanceOf()
    return "ClassStandardCrashbreaker"
end

-------------------------------------------------------------------------------------------------------
--Overridden Methods
--For method documentation, see superclass
----------------------------------------------------------------------------------------------------

function ClassStandardCrashbreaker:unlock()
    ClassCrashbreaker.unlock(self)

    guihooks.trigger("i_messages_enqueue", {
        text = "CRASHBREAKER READY",
        cssClass = "flame",
        relPath = "pickups/crashbreaker.png",
        duration = 5,
        colour = "Yellow"
    })
end

function ClassStandardCrashbreaker:use(vehicle)
    --[[
    Begins the usage sequence
    Parameters:
        vehicle - the vehicle triggering the crashbreaker
    ]]--
    if not self:canBeUsed() then
        return
    end

    errorHandler.assertNil(vehicle)

    self.inUse = true

    --[[Set the explosive's position at the vehicle's
        Since position is a table, immutability requires a cloned vector]]--
    self.useData.position = vehicle:getPosition():clone()
    self.useData.vehicleTrigger = vehicle:getId()

    guihooks.trigger("i_crashbreaker_update", {used = true})
    soundHandler.play("crashbreaker_explosion")

    self.shockwave:begin(self.useData.position)

    --Destroy the vehicle
    vehicle:breakAll()
end

function ClassStandardCrashbreaker:reset()
    ClassCrashbreaker.reset(self)
    self.shockwave:reset()

    self.useData = {
        timer = 0,
        vehicleTrigger = nil,
        position = nil
    }

    for k, _  in pairs(self.vehicles) do
        self.vehiclesAffected[k] = false
    end
end

-------------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassStandardCrashbreaker:new(vehicles, vehicleRequirement)
    --[[
    See ClassCrashbreaker:new()
    ]]--
    local self = ClassCrashbreaker:new(vehicles, vehicleRequirement)

    --Stores whether a vehicle has been affected
    self.vehiclesAffected = {}
    for k, _  in pairs(self.vehicles) do
        self.vehiclesAffected[k] = false
    end

    self.constants = {
        --The diameter the explosive affects other objects
        EXPLOSION_RANGE = 40,
        --The total time the explosive lasts
        PERIOD = 1,
        --The maximum alpha of the shockwaves
        MAX_ALPHA = 1,
        --The number of shockwaves
        NUM_SHOCKWAVES = 5
    }

    self.useData = {
        timer = 0,
        position = nil,
        vehicleTrigger = nil
    }

    self.shockwave = ClassShockwave:new(
            "crashbreaker_shockwave_",
            self.constants.NUM_SHOCKWAVES,
            self.constants.EXPLOSION_RANGE,
            self.constants.PERIOD,
            function(time, alpha)
                return "1 "..(1 - time/self.constants.NUM_SHOCKWAVES).." 0 "..alpha
            end,
            nil
    )

    setmetatable(self, {__index = ClassStandardCrashbreaker})
    return self
end

------------------------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassStandardCrashbreaker