----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local classHelper  = require("scenario/intersection/classHelper")

------------------------------------------------------------------------------------------------------------------

--[[
NOTE: This class is implied as an ABSTRACT, so no instances should be created!
]]--

local ClassPickup = classHelper.inheritsFrom(nil)

----------------------------------------------------------------------------------------------------
--Static Attributes
----------------------------------------------------------------------------------------------------

ClassPickup.className = "ClassPickup"

------------------------------------------------------------------------------------------------------------------
--Instance Methods
----------------------------------------------------------------------------------------------------

function ClassPickup:getType()     return self.type      end
function ClassPickup:getValue()    return self.value     end
function ClassPickup:getPosition() return self.position  end
function ClassPickup:isCollected() return self.collected end

function ClassPickup:getPickupRange()
    --[[
    Returns:
        <number> - the maximum distance at which the pickup can be collected
    ]]--
    return 4
end

function ClassPickup:collect()
    --[[
    Collects the pickup by removing from the world and altering the UI
    ]]--
    self.collected = true
    guihooks.trigger("i_pickup_add", {pickup = self.type})
end

function ClassPickup:reset()
    --[[
    Resets the pickup by re-showing if it isn't already and resetting back
    to the initial state
    ]]--
    self.collected = false
    self.scale = self.INIT_SCALE
    self.anim.speed = self.anim.INIT_SPEED
    self.anim.angle = 0

    TorqueScript.eval(self.type..'.scale = "'..self.scale..' '..self.scale..' '..self.scale..'";')
end

function ClassPickup:update(dt)
    --[[
    Updates the pickup by animating it
    ]]--
    local anim = self.anim
    anim.angle = anim.angle + anim.speed*dt
    anim.angle = anim.angle % 360

    TorqueScript.eval(self.type..'.rotation = "0 0 -1 '..anim.angle..'";')

    if self.collected then
        --The speed function
        anim.speed = anim.INIT_SPEED * (1 - self.scale/self.INIT_SCALE) * anim.COLLECT_SPEED_MULTI

        if self.scale > 0 then
            TorqueScript.eval(self.type..'.scale = "'..self.scale..' '..self.scale..' '..self.scale..'";')
            self.scale = self.scale - self.anim.SCALE_DEC*dt
        elseif self.scale < 0 then
            TorqueScript.eval(self.type..'.scale = "0 0 0";')
            --Prevents the scale being continuously set
            self.scale = 0
        end
    end
end

----------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassPickup:new(position, type)
    --[[
    Parameters:
        position - the position in-world
        type     - the particular type of the pickup
    ]]--
    errorHandler.assertNil(position, type)

    local self = {}
    self.position = position
    self.type = type
    --Should be overridden when inherited
    self.value = nil
    self.collected = false

    self.anim = {
        angle = 0,
        --The speed of rotation (degree/s)
        INIT_SPEED = 90,
        --The maximum speed mutliplier to apply to the collection animation
        COLLECT_SPEED_MULTI = 30,
        --Scale decretion /s
        SCALE_DEC = 4
    }
    self.anim.speed = self.anim.INIT_SPEED

    self.INIT_SCALE = scenetree.findObject(type):getScale().z
    self.scale = self.INIT_SCALE

    setmetatable(self, {__index = ClassPickup})
    return self
end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassPickup