----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local ClassVehicle   = require("scenario/intersection/classes/vehicles/ClassVehicle")
local ClassShockwave = require("scenario/intersection/classes/classShockwave")
local ClassTimer     = require("scenario/intersection/classes/classTimer")
local classHelper    = require("scenario/intersection/classHelper")
local errorHandler   = require("scenario/intersection/handlers/errorHandler")

----------------------------------------------------------------------------------------------------

local ClassAi = classHelper.inheritsFrom(ClassVehicle)

----------------------------------------------------------------------------------------------------
--Static Attributes
----------------------------------------------------------------------------------------------------

ClassAi.className = "ClassAi"
ClassAi.scenarioName = ""
ClassAi.pathDir = ""

--The maximum points for damages
ClassAi.MAX_DAMAGE_SCORE = 3000
--The ratio of damage : points
ClassAi.DAMAGE_CUT = 0.15

----------------------------------------------------------------------------------------------------
--Static Methods
----------------------------------------------------------------------------------------------------

function ClassAi:setScenarioName(scName)
    --[[
    Sets the scenario name, and the path base directory
    Parameters:
        scName - the scenario name
    ]]--
    errorHandler.assertNil(scName)

    ClassAi.scenarioName = scName
    ClassAi.pathDir = "/replays/scriptai/tracks/"..scName
end

----------------------------------------------------------------------------------------------------
--Instance Methods
----------------------------------------------------------------------------------------------------

function ClassAi:stopPathing()
    --[[
    Ends the ai pathing for this vehicle
    ]]--
    self.obj:queueLuaCommand('ai.stopFollowing(true, false)')
end

function ClassAi:startPathing()
    --[[
    Begins the ai vehicle on its path
    ]]--
    if self.stationary then
        return
    end

    local data = jsonReadFile(ClassAi.pathDir.."/"..self.name..".track.json")

    --Don't loop
    self.obj:queueLuaCommand('ai.startFollowing('..serialize(data.recording)..', nil, 1, "noLoop")')
end

function ClassAi:setHazards(state)
    --[[
    Turn the hazards on/off
    Parameters:
        state - boolean of whether the hazards are on or off
    ]]--
    self.obj:queueLuaCommand('electrics.set_warn_signal('..tostring(state)..')')
end

function ClassAi:setActive()
    --[[
    Starts the active state of the vehicle, such as pathing
    ]]--
    --Don't move a stationary vehicle
    if not self.stationary then
        self:teleportTo(self.initPosition)
        self:startPathing()
        self.shockwave:begin(self.initPosition)
    end
end

function ClassAi:pend()
    self.pendStart = true

    if self.timer then
        self.timer:setActive(true)
    else
        --Timer doesn't exists so show immediately
        self:setActive()
    end
end

function ClassAi:isEliminated() return self.eliminated end
function ClassAi:isStationary() return self.stationary end

----------------------------------------------------------------------------------------------------
--Overridden Methods
----------------------------------------------------------------------------------------------------

function ClassAi:getDamageScore()
    --Vehicle has been eliminated so ignore
    if self.eliminated then
        return 0
    end

    return ClassVehicle.getDamageScore(self)
end

function ClassAi:setCrashed()
    if self.crashed then
        return
    end

    self.crashed = true
    --Update the UI
    guihooks.trigger("i_scoring_add_vehicle")
end

function ClassAi:eliminate()
    self.elimShockwave:begin(self.position:clone())

    --Set Ai to some position out of view and effect
    local pos = self.position:clone();
    pos:setY(pos:getY() + 10000);
    self:teleportTo(pos, true)

    self:lock()
    self:stopPathing()
    self.eliminated = true

    return false
end

function ClassAi:lock()
    ClassVehicle.lock(self)
    self.pendStart = false
end

function ClassAi:setHealthbar(colour, scale, rotation)
    --[[
    Sets the colour and scale of the healthbar
    Parameters:
        colour   - the colour to set (can be nil to skip)
        scale    - the scale of the healthbar
        rotation - the rotation of the healthbar
    ]]--
    errorHandler.assertNil(scale, rotation)
    errorHandler.assertTrue(scale >= 0 and scale <= self.HEALTHBAR_LENGTH, "Scale needs to be >= 0 and <= the maximum length ["..scale.."]")

    local name = self.name.."_healthbar"

    if colour then
        --Recreate with new colour
        if scenetree.findObject(name) then
            TorqueScript.eval(name..".delete();")
        end

        TorqueScript.eval([[
          new TSStatic(]]..name..[[) {
            shapeName = "levels/smallgrid/art/intersection/vehicles/healthbar/]]..colour..[[.dae";
            dynamic = "1";
            scale = "1 1 1";
          };
        ]])
    end

    local obj = scenetree.findObject(name)
    obj:setField("rotation", 0, "0 0 1 "..math.deg(rotation))
    obj:setField("position", 0, self.center:getX().." "..self.center:getY().." "..(self.center:getZ()+self.vehicleData.height))
end

function ClassAi:updateHealthbar()
    --[[
    Updates the scale and colour of the healthbar
    ]]--
    local percDamage = 1 - (self:getDamageScore() / ClassAi.MAX_DAMAGE_SCORE)

    --Work out the colour
    local colour = ""
    if     percDamage < 0.25 then colour = "red"
    elseif percDamage < 0.5  then colour = "orange"
    elseif percDamage < 0.75 then colour = "yellow"
    else                          colour = "green"
    end

    --Work out the scale
    local scale = self.HEALTHBAR_LENGTH * percDamage

    local healthbarObj = scenetree.findObject(self.name.."_healthbar")
    if not healthbarObj.shapeName:find(colour) then
        --Set the colour and scale as it is different
        self:setHealthbar(colour, scale, self.rotationX)
    else
        --Just ste the scale, as colour is the same
        self:setHealthbar(nil, scale, self.rotationX)
    end
end

function ClassAi:update(dt)
    ClassVehicle.update(self, dt)

    if self.pendStart and self.timer then
        self.timer:update(dt)
    end

    if not self.shockwave:hasFinished() then
        self.shockwave:update(dt)
    end

    if not self.elimShockwave:hasFinished() then
        self.elimShockwave:update(dt)
    end

    --self:updateHealthbar()
end

function ClassAi:teleportTo(position)
    --[[
    Teleports the vehicle to a location, with or without a visual effect
    Parameters:
        position - the position to send the vehicle to
    ]]--
    TorqueScript.eval(self.name..'.position = "'..position:getX()..' '..position:getY()..' '..position:getZ()..'";')

    ClassVehicle.update(self, nil)
end

function ClassAi:initPos()
    --[[
    Sets the vehicle away from view, unless stationary
    ]]--
    if not self.stationary then
        local pos = self.position:clone()
        pos:setZ(0.5)
        self:teleportTo(pos)
    end
end

function ClassAi:reset()
    ClassVehicle.reset(self)

    self:stopPathing()
    self.eliminated = false
    self.pendStart = false
    self:initPos()
    self:setHazards(false)

    if self.timer then
        self.timer:reset()
    end

    if self.stationary then
        self.tracking = true
    end

    self.shockwave:reset()
    self.elimShockwave:reset()
end

----------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassAi:new(id, name, position, delay)
    --[[
    See ClassVehicle:new()
    Additional parameters:
        delay - the delay between the scenario starting and the AI pathing
    ]]--
    local self = ClassVehicle:new(id, name, position, delay)

    self.initPosition = self.position:clone()
    self.eliminated = false
    self.pendStart = false

    self.shockwave = ClassShockwave:new(
            self.name.."_entryShockwave",
            3,
            15,
            2,
            function(_, alpha)
                return "0 1 0 "..alpha
            end,
            function()
                self.tracking = true
                self.shockwave:reset()
            end
    )

    self.elimShockwave = ClassShockwave:new(
            self.name.."exitShockwave",
            3,
            15,
            2,
            function(_, alpha)
                return "1 0 0 "..alpha
            end,
            function()
                self.elimShockwave:reset()
            end
    )

    self.timer = nil
    if delay > 0 then
        --If specified, set the delay and set the callback to be the teleport action
        self.timer = ClassTimer:new(
                delay,
                true,
                function()
                    self:setActive()
                end
        )
    end

    self.HEALTHBAR_LENGTH = self.vehicleData.length * 5

    setmetatable(self, {__index = ClassAi})

    --If there is no path, then it must be stationary
    self.stationary = not FS:fileExists(ClassAi.pathDir.."/"..self.name..".track.json")

    --Vehicle won't move, so set tracking instantly
    if self.stationary then
        self.tracking = true
    end

    self:initPos()

    return self
end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassAi
