-------------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")
local ClassVector  = require("scenario/intersection/classes/classVector")
local classHelper  = require("scenario/intersection/classHelper")

----------------------------------------------------------------------------------------------------

--[[
NOTE: This class is implied as ABSTRACT, so no instances should be created!
]]--

local ClassVehicle = classHelper.inheritsFrom(nil)

-------------------------------------------------------------------------------------------------------
--Static Attributes
----------------------------------------------------------------------------------------------------

ClassVehicle.className = "ClassVehicle"
ClassVehicle.GRAV_CONST = 6.67*10^-11

-------------------------------------------------------------------------------------------------------
--Static Methods
----------------------------------------------------------------------------------------------------

function ClassVehicle:getVehicleData(vehObj)
    --[[
    Returns a table of vehicle specific attributes
    Parameters:
        vehObj - the vehicle object
    Returns:
        <table> - the table of points, contains:
            height - the height of the vehicle
            center - the center point of the vehicle
            length - the length of the vehicle
    ]]--
    local data = {
        center = vec3(0,0,0),
        height = 0,
        length = {
            min = {x =  math.huge, y =  math.huge},
            max = {x = -math.huge, y = -math.huge}
        }
    }

    --Simplify accessing
    local len = data.length

    local nodeCount = vehObj:getNodeCount()

    for i = 0, nodeCount do
        local pos = vehObj:getNodePosition(i)

        --Update center point
        data.center = data.center + pos

        --Update height
        if pos.z > data.height then
            data.height = pos.z
        end


        --Update min/max points
        if pos.y < len.min.y then
            len.min.y = pos.y
        elseif pos.y > len.max.y then
            len.max.y = pos.y
        end

        if pos.x < len.min.x then
            len.min.x = pos.x
        elseif pos.x > len.max.x then
            len.max.x = pos.x
        end
    end

    --Sum / total nodes
    data.center = data.center / nodeCount

    --Set the length to the differential of max and min (finding the longest axis)
    data.length = math.max(len.max.y - len.min.y, len.max.x - len.min.x)

    return data
end

--function ClassVehicle:getRollPitch(dirFront, dirUp)
--    --[[
--    Returns the roll and pitch of a vehicle
--    This function is simply a copy-paste of the vanilla BeamNG one, with a few changes
--    ]]--
--    -- find vehicle roll and pitch, in degrees, 0deg being normal upright rotation, +/-180deg being on its roof
--    local dirLeft = dirUp:cross(dirFront)
--    local roll  = math.deg(math.asin(dirLeft.z))
--    local pitch = math.deg(math.asin(dirFront.z))
--    if dirUp.z < 0 then -- if we are closer to upside down than to downside up
--        -- detect the "on its roof" situation, where angles are zero, and make sure they go all the way to 180deg instead, like this:
--        -- original rotation angles:  0deg (ok), 90deg (halfway),      0deg (on its roof), -90deg (halfway), 0deg (ok)
--        -- corrected rotation angles: 0deg (ok), 90deg (halfway), +/-180deg (on its roof), -90deg (halfway), 0deg (ok)
--        roll  = sign( roll)*(180 - math.abs( roll))
--        pitch = sign(pitch)*(180 - math.abs(pitch))
--    end
--
--    return roll, pitch
--end
--
--function ClassVehicle:getInitRotation(vehObj)
--    --[[
--    Returns the vehicle rotation around the z-axis
--    This function is simply a copy-paste of the vanilla BeamNG one, with a few changes
--    ]]--
--    local rot
--    local dirUp = vec3(vehObj:getDirectionVectorUp())
--    local dirFront = vec3(vehObj:getDirectionVector())
--    local roll, pitch = ClassVehicle:getRollPitch(dirFront, dirUp)
--
--    if math.abs(pitch) > 80 or math.abs(roll) > 45 then
--        rot = quatFromDir(-dirFront, vec3(0,0,1))
--    else
--        rot = quatFromDir(-dirFront, dirUp)
--    end
--
--    rot = rot:toEulerYXZ()
--
--    return rot.z
--end

function ClassVehicle:getInitRotationX(vehObj)
    --[[
    Returns:
        <number> - the angle around the x-axis in radians
    ]]--
    local rot = vehObj:getRotation()
    --Reconstuct a quaternion from userdata
    local q = quat(rot.x, rot.y, rot.z, rot.w)
    local euler = q:toEulerYXZ()

    return euler.x
end

----------------------------------------------------------------------------------------------------
--Instance Methods
----------------------------------------------------------------------------------------------------

function ClassVehicle:getId()          return self.id           end
function ClassVehicle:getObj()         return self.obj          end
function ClassVehicle:getName()        return self.name         end
function ClassVehicle:hasCrashed()     return self.crashed      end
function ClassVehicle:hasIgnited()     return self.ignited      end
function ClassVehicle:getPosition()    return self.position     end
function ClassVehicle:isAIControlled() return error("isAIControlled() must be overridden in sub class") end

function ClassVehicle:getObjData() return map.objects[self.id] end
function ClassVehicle:getVelocity() return self:getObjData().vel end

function ClassVehicle:getPendingAccel()
    --[[
    Returns:
        <ClassVector> - the sum vector of the acceleration pending
    ]]--
    local accel = ClassVector:new(0,0,0)

    for _, accelData in ipairs(self.forcedAccel) do
        accel = accel:add(accelData.a)
    end

    return accel
end

function ClassVehicle:getDamageScore()
    --[[
    Returns:
        <number> - the score from the damage caused
    ]]--
    return math.floor(math.min(self.MAX_DAMAGE_SCORE, self:getObjData().damage * self.DAMAGE_CUT))
end

function ClassVehicle:setCrashed() self.crashed = true end

function ClassVehicle:trackDamage()
    --[[
    Tracks the damage to the vehicle and allows calls by onObjectCollided
    ]]--
    scenario_scenarios.trackVehicleMovementAfterDamage(self.name)
end

function ClassVehicle:lock()
    --[[
    Locks the car to prevent continued driving
    ]]--
    self.obj:queueLuaCommand('controller.setFreeze(1)')
end

function ClassVehicle:ignite()
    --[[
    Sets the vehicle on fire
    ]]--
    self.obj:queueLuaCommand("fire.explodeVehicle()")
    self.ignited = true
end

function ClassVehicle:breakAll()
    --[[
    Sets the vehicle on fire
    ]]--
    self.obj:queueLuaCommand("beamstate.breakAllBreakgroups()")
end

function ClassVehicle:boost(acceleration, position, time)
    --[[
    Pushes the vehicle in a direction with a given force using a planet
    Parameters:
        acceleration - the magnitude of the acceleration to set as initial
        position     - the position to set the source of the force at
        time         - the amount of time to apply the force for
    ]]--
    errorHandler.assertNil(acceleration, position, time)
    --[[
    The planets in-built system uses GM/r² -> a = GM/r²
    To apply an acceleration a, we need to find M = ar²/G
    Then, set M to be negative to repel the vehicle
    ]]--
    local distance = self.position:getDistanceBetween(position)
    local mass = -acceleration / ClassVehicle.GRAV_CONST * distance^2

    self.obj:queueLuaCommand("beamstate.addPlanet(vec3("..position:getX()..", "..position:getY()..", "..position:getZ().."), 1, "..mass..", "..time..")")
end

function ClassVehicle:eliminate()
    --[[
    Eliminates this vehicle from the scenario
    If the player, returns true to indicate scenario failure
    Else         , teleports the vehicle away
    Returns:
        <boolean> - whether the scenario should fail or not
    ]]--
    error("eliminate() must be overridden in sub class")
end

--function ClassVehicle:updateRotationX(prevPos)
--    --[[
--    Updates the rotationX value
--    Parameters:
--        prevPos - the previous position (before update)
--    ]]--
--    local difX = prevPos:getX() - self.position:getX()
--    local difY = prevPos:getY() - self.position:getY()
--
--    --Angle hasn't changed
--    if difY == 0 then
--        return
--    end
--
--    self.rotationX = math.atan(difX / difY)
--end

function ClassVehicle:update(_)
    --[[
    Updates the vehicle such as updating its position internally
    ]]--
    local newPos = self.obj:getPosition()

    self.position:setX(newPos.x)
    self.position:setY(newPos.y)
    self.position:setZ(newPos.z)

    --Update center position
    self.center = ClassVector:new(
            self.position:getX() + self.vehicleData.center.x,
            self.position:getY() + self.vehicleData.center.y,
            self.position:getZ() + self.vehicleData.center.z
    )
end

function ClassVehicle:reset()
    --[[
    Resets the instance to the initial state
    ]]--
    self.ignited = false
    self.crashed = false
    self.tracking = false
    self.acceleration = ClassVector:new(0,0,0)
    self:trackDamage()
end

function ClassVehicle:instanceOf()
    return "ClassVehicle"
end

----------------------------------------------------------------------------------------------------
--Constructor
----------------------------------------------------------------------------------------------------

function ClassVehicle:new(id, name, position)
    --[[
    Parameters:
        id       - the in-world id
        name     - the name in-world
        position - its initial position
    ]]--
    errorHandler.assertNil(id, name, position)

    local self = {}
    self.id = id
    self.name = name
    self.position = position
    self.obj = scenetree.findObject(name)
    self.crashed = false
    self.ignited = false
    self.forcedAccel = {}
    self.acceleration = ClassVector:new(0,0,0)
    self.tracking = false

    setmetatable(self, {__index = ClassVehicle})

    self.vehicleData = self:getVehicleData(self.obj)

    local relCenter = self.vehicleData.center
    self.center = ClassVector:new(
            self.position:getX() + relCenter.x,
            self.position:getY() + relCenter.y,
            self.position:getZ() + relCenter.z
    )

    return self
end


----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

return ClassVehicle