local M = {}

----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

--Classes
local ClassAi = require("scenario/intersection/classes/vehicles/classAi")

--Handlers
local pickupHandler  = require("scenario/intersection/handlers/pickupHandler")
local scoringHandler = require("scenario/intersection/handlers/scoringHandler")
local vehicleHandler = require("scenario/intersection/handlers/vehicleHandler")
local errorHandler   = require("scenario/intersection/handlers/errorHandler")
local timerHandler   = require("scenario/intersection/handlers/timerHandler")
local fileHandler    = require("scenario/intersection/handlers/fileHandler")

--Miscellaneous
local tableComp = require("scenario/intersection/tableComprehension")


----------------------------------------------------------------------------------------------------
--Variables
----------------------------------------------------------------------------------------------------

local sourceFile = ""
local bullettimeCurr = 1

----------------------------------------------------------------------------------------------------

local function endScenario(completed, reason)
    --[[
    Triggers the ending of the scenario
    Parameters:
        completed - whether the scenario has been completed (ie. not terminated early)
        reason    - the reason if the scenario was terminated early (optional)
    ]]--
    errorHandler.assertNil(completed)

    timerHandler.setScenarioState("finished")

    local vehicles = vehicleHandler.getVehicles()
    local pickups = pickupHandler.getPickups()

    local results, multipliers, target = scoringHandler.formatToUI(pickups.coins, pickups.multipliers, vehicles)

    local data = {
        success = completed,
        reason = reason,
        name = ClassAi.scenarioName:gsub("_"," "),
        results = results
    }

    --May not always be present in the scenario
    if multipliers then
        data.multipliers = multipliers
    end
    if target then
        data.target = target
    end

    --Handle base medals and scores
    local scenario = jsonDecode(readFile(sourceFile), sourceFile)[1]
    local scores = scenario.intersection.medalRequirement

    local save = fileHandler.readWholeFile().scenarios
    local scData = save[ClassAi.scenarioName]

    --Check it exists and create it if it doesn't
    if not scData then
        fileHandler.addScenario(ClassAi.scenarioName)
        scData = fileHandler.readWholeFile().scenarios[ClassAi.scenarioName]
    end

    --Best
    data.medalData = {best = {
        score = scData.score,
        gotten = scData.score < results[#results].amount}}

    --Generic
    for k, score in pairs(scores) do
        data.medalData[k] = {score = score, gotten = score <= scData.score and scData.score > 0}
    end


    guihooks.trigger("i_endScreen", data)
end

local function setupScenario()
    --[[
    Sets up the scenario by creating all instances and setting the timer state to 'ready'
    ]]--
    local scenario = jsonDecode(readFile(sourceFile), sourceFile)[1]
    local config = scenario.intersection

    ClassAi:setScenarioName(sourceFile:gsub("/levels/smallgrid/scenarios/intersection/", ""):gsub(".json", ""))

    --Create instances and let them know of any delay and target vehicle
    vehicleHandler.createVehicleInstances(config.delay, config.targetVehicle)

    local vehicles = vehicleHandler.getVehicles()
    local vehs = tableComp.merge(vehicles.ais, vehicles.props)
    vehs[vehicles.player:getId()] = vehicles.player
    pickupHandler.setupCrashbreaker(vehs, config.crashbreakerRequirement)

    pickupHandler.createPickupInstances()

    timerHandler.setScenarioState("ready")
end

local function resetScenario()
    --[[
    Resets the scenario back to the setup state for the next attempt
    ]]--
    --Reset handlers
    vehicleHandler.reset()
    pickupHandler .reset()
    scoringHandler.reset()
end

local function resetUI()
    --[[
    Resets the UI for the next attempt
    ]]--
    guihooks.trigger("i_scoring_init", {vehicles = tableComp.lengthOfTable(vehicleHandler.getVehicles().ais)})
end

local function resetState()
    --[[
    Resets the states of each handler to their initial state
    ]]--
    timerHandler  .resetState()
    pickupHandler .resetState()
    vehicleHandler.resetState()
    scoringHandler.resetState()
end

----------------------------------------------------------------------------------------------------
--BeamNG Functions
----------------------------------------------------------------------------------------------------

local function onPreRender(dt)
    local bt = bullettime.get()

    --Update the UI
    if bt ~= bullettimeCurr then
        bullettimeCurr = bt
        guihooks.trigger("i_bullettime_changed", {speed = bt})
    end

    --Stop updating when scenario is paused
    if bt == 0 then
        return
    end

    local dtScaled = dt * bt
    local dtTime = dt

    --Update the timer
    local finished = timerHandler.update(dt, dtScaled)

    local timerState = timerHandler.getScenarioState()
    local crashbreaker = pickupHandler.getCrashbreaker()

    --[[During the running state, the time is slowed down during slow motion
        This is not true for the timeout/pre-timeout states]]--
    if timerState == "running" then
        dtTime = dtScaled
    end

    if timerState == "running" or timerState == "timeout" or timerState == "pre-timeout" then
        local player = vehicleHandler.getVehicles().player
        --Get the value of the hazards
        player:getObj():queueLuaCommand("obj:queueGameEngineLua('cbActivate = '..electrics.values.lowhighbeam)")

        --Check to see if the crashbreaker should be activated
        if cbActivate == 1 and crashbreaker:canBeUsed() then
            crashbreaker:use(player)
            timerHandler.initCrashtime()
        end

        --Update vehicles
        vehicleHandler.update(dt)
        --Update pickup collection
        pickupHandler.checkCollectables(player)
    end

    if timerState ~= "init" and crashbreaker then
        if crashbreaker:beingUsed() then
            crashbreaker:update(dtScaled)
        end
    end

    pickupHandler.update(dtScaled)

    if finished then
        endScenario(true)
    end
end

local function onObjectCollision(id1, id2)
    local crashbreaker = pickupHandler.getCrashbreaker()
    --Find the two vehicle instances with the respective id's
    local instance1, instance2 = vehicleHandler.getVehicle(id1), vehicleHandler.getVehicle(id2)

    local areAIs     = instance1:instanceOf("ClassAi")   and instance2:instanceOf("ClassAi")
    local areProps   = instance1:instanceOf("ClassProp") and instance2:instanceOf("ClassProp")
    local arePathing = not instance1:hasCrashed() and not instance2:hasCrashed()

    --[[Ignore:
        - Two AIs crashing that are still on their path
        - Two props crashing into each other]]--
    if (areAIs and arePathing) or areProps then
        return
    end


    --Add the scoring
    scoringHandler.addScoreFromCollision(instance1, instance2)


    --Update the vehicles themselves
    for _, instance in ipairs({instance1, instance2}) do
        local isAi = instance:instanceOf("ClassAi")

        timerHandler.checkForCrashtime(vehicleHandler.getVehicles().player, instance1, instance2)

        if not instance:hasCrashed() then
            --Subtract from the requirement
            if isAi then
                crashbreaker:addCrashedVehicle()
            end

            instance:setCrashed()
        end

        if isAi then
            --Any Ais should stop following their path and show hazards
            instance:stopPathing()
            instance:setHazards(true)
        end

        --Vehicles should be locked down to prevent moving
        instance:lock()
    end
end

local function onBeamNGTrigger(data)
    if data.event == "enter" then
        if data.triggerName:find("outBounds") then
            local vehicle = vehicleHandler.getVehicleByName(data.subjectName)

            --Eliminate AIs that have made it to the exit while still pathing
            if vehicle:instanceOf("ClassAi") and not vehicle:hasCrashed() then
                vehicle:eliminate()
            elseif vehicle:instanceOf("ClassPlayer") then
                --End scenario
                endScenario(false, "Eliminated")
            end
        end
    end
end

local function onRaceStart()
    timerHandler.setScenarioState("running")
    vehicleHandler.start()
end

local function onCountdownStarted()
    guihooks.trigger("i_countdown_start")
end

local function onScenarioRestarted()
    guihooks.trigger("i_reset")

    timerHandler.reset()
end

local function onScenarioChange(sc)
    if sc and sc.state == "pre-running" then
        local srcFile = sc.sourceFile

        local isSameSc = sourceFile == srcFile

        --[[Since this module will retain its variables from the previous scenario, re-setup the scenario
            for correct loading of a new scenario]]--
        if not isSameSc then
            --Not the first scenario so reset variables that are retained (unecessary for first scenario)
            if sourceFile then
                resetState()
            end

            sourceFile = srcFile
            setupScenario()
        else
            resetScenario()
        end

        --UI needs to be reset after pre-running occurs or it will be wiped by scenario
        resetUI()
    end
end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

M.onPreRender         = onPreRender
M.onObjectCollision   = onObjectCollision
M.onBeamNGTrigger     = onBeamNGTrigger
M.onRaceStart         = onRaceStart
M.onCountdownStarted  = onCountdownStarted
M.onScenarioRestarted = onScenarioRestarted
M.onScenarioChange    = onScenarioChange
return M
