local M = {}

----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

local errorHandler = require("scenario/intersection/handlers/errorHandler")

----------------------------------------------------------------------------------------------------

local function getDefaultScenarioData()
    --[[
    Returns a blank set of data for a scenario in the savefile
    Returns:
        <table> - the data for a scenario
    ]]--
    return {
        score = 0,
        medal = "",
        stats = {
            crashedAll = false
        }
    }
end

local function createNewSavefile()
    --[[
    Creates a brand new savefile
    ]]--
    local fileDirectory = "/levels/smallgrid/scenarios/intersection/"
    local files = FS:findFilesByPattern(fileDirectory, "*.json", 1, true, false)

    --Structure: savefile - hubworld, scenarios
    local content = {hubworld = {}, scenarios = {}}

    --Add each scenario into the savefile
    for _, fileName in pairs(files) do
        if not fileName:find("hubworld") then
            local scenarioName = fileName:gsub(fileDirectory, ""):gsub(".json", "")
            content.scenarios[scenarioName] = getDefaultScenarioData()
        end
    end

    serializeJsonToFile("intersection_savefile", content, true)
end

local function checkFileExists()
    --[[
    Checks to see if it can open the savefile. If it can then also return the file
    so any function calling this doesn't have to open the file again - efficiency
    Returns:
        <boolean> - whether the file exists
        <table>   - the file
    ]]--
    local file = readFile("intersection_savefile")
    if file then
        return true, file
    else
        return false
    end
end

local function readWholeFile()
    --[[
    Checks the file exists and it up to date, and then reads the file
    Returns:
        <table> - the save file
    ]]--
    local fileExists, file = checkFileExists()
    if not fileExists then
        createNewSavefile()
        --Update the file variable with newly created info
        _, file = checkFileExists()
    end

    return jsonDecode(file, "intersection_savefile")
end

local function addScenario(scName)
    --[[
    Adds a new scenario to the savefile
    Parameters:
        scName - the scenario to add
    ]]--
    local save = readWholeFile()
    save.scenarios[scName] = getDefaultScenarioData()

    serializeJsonToFile("intersection_savefile", save, true)
end

local function saveToFile(scName, scData)
    --[[
    Saves the data for a scenario to the savefile
    Parameters:
        scName - the scenario name
        scData - the results data for the scenario
    ]]--
    errorHandler.assertNil(scName, scData)
    errorHandler.assertTrue(scData.medal ~= nil and scData.score ~= nil, "scData does not correct information")

    local save = readWholeFile()
    if scName == "hubworld" then
        --TODO do something with hubworld here
    else
        local scSave = save.scenarios[scName]
        scSave.medal = scData.medal
        scSave.score = scData.score
    end

    serializeJsonToFile("intersection_savefile", save, true)
end


M.readWholeFile = readWholeFile
M.addScenario = addScenario
M.saveToFile = saveToFile
M.createNewSavefile = createNewSavefile
return M