local M = {}

----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

--Classes
local ClassCoin                 = require("scenario/intersection/classes/pickups/classCoin")
local ClassVector               = require("scenario/intersection/classes/classVector")
local ClassMultiplier           = require("scenario/intersection/classes/pickups/classMultiplier")
local ClassArmageddon           = require("scenario/intersection/classes/pickups/classArmageddonPickup")
local ClassStandardCrashbreaker = require("scenario/intersection/classes/crashbreaker/classStandard")

--Handlers
local errorHandler = require("scenario/intersection/handlers/errorHandler")

----------------------------------------------------------------------------------------------------
--Variables
----------------------------------------------------------------------------------------------------

local pickups = {coins = {}, multipliers = {}, explosives = {}}
local crashbreaker

----------------------------------------------------------------------------------------------------

local function findPickups()
    --[[
    Finds all pickups in the scenario
    ]]--
    local found = {coins = {}, multipliers = {}, explosives = {}}

    for i, value in ipairs(ClassCoin:getTypes()) do
        local coin = scenetree.findObject(value)
        --All coins MUST be in the scenario
        errorHandler.assertTrue(coin ~= nil, value.." is missing in this scenario")

        found.coins[i] = {value = value, class = "coin", position = coin:getPosition()}
    end

    for _, value in ipairs(ClassMultiplier:getTypes()) do
        local multi = scenetree.findObject(value)
        --Not all types of multiplier may be in the scenario
        if multi then
            table.insert(found.multipliers, {value = value, class = "multiplier", position = multi:getPosition()})
        end
    end

    local armageddon = scenetree.findObject("armageddon")
    if armageddon then
        table.insert(found.explosives, {value = "armageddon", class = "armageddon", position = armageddon:getPosition()})
    end

    return found
end

local function createPickupInstances()
    --[[
    Creates instances of ClassPickup (ClassPickup is implied as abstract so only those that are
    sub-classes are actually instantiated) from those in-world
    ]]--
    local foundPickups = findPickups()

    for type, found in pairs(foundPickups) do
        for _, data in ipairs(found) do
            local value = data.value
            local class = data.class
            local pos = data.position

            local position = ClassVector:new(pos.x, pos.y, pos.z)

            local instance

            --Create respective instance
            if class == "coin" then
                instance = ClassCoin:new(position, value)
            elseif class == "multiplier" then
                instance = ClassMultiplier:new(position, value)
            elseif class == "armageddon" then
                instance = ClassArmageddon:new(position, crashbreaker)
            else
                error((class or "nil").." is not a valid sub-class of ClassPickup")
            end

            pickups[type][value] = instance
        end
    end
end

local function checkCollectables(vehicle)
    --[[
    Checks each pickup to see if this vehicle is close enough to collect them
    If there are, this function also collects them
    Parameters:
      vehicle - the vehicle to check
    ]]--
    errorHandler.assertNil(vehicle)

    for _, v in pairs(pickups) do
        for _, instance in pairs(v) do
            --No point in checking those that are collected
            if not instance:isCollected() then
                local dist = vehicle:getPosition():getDistanceBetween(instance:getPosition())
                if dist <= instance:getPickupRange() then
                    if instance:instanceOf("ClassArmageddonPickup") then
                        crashbreaker = instance:collect(crashbreaker)
                    else
                        instance:collect()
                    end
                end
            end
        end
    end
end

local function resetPickups()
    --[[
    Resets the pickups
    ]]--
    --Simply reset, rather than re-creating
    for _, table in pairs(pickups) do
        for _, instance in pairs(table) do
            instance:reset()
        end
    end
end

local function update(dt)
    --[[
    Updates pickups
    Parameters:
        dt - the time since the last frame
    ]]--
    for _, table in pairs(pickups) do
        for _, instance in pairs(table) do
            instance:update(dt)
        end
    end
end

local function setupCrashbreaker(vehicles, requirement)
    --[[
    Creates the crashbreaker
    See ClassStandardCrashbreaker:new()
    ]]--
    crashbreaker = ClassStandardCrashbreaker:new(vehicles, requirement)
end

local function resetCrashbreaker()
    --[[
    Resets the crashbreaker
    ]]--
    crashbreaker:reset()

    if not crashbreaker:instanceOf("ClassStandardCrashbreaker") then
        --Convert the class back to the standard version
        crashbreaker = ClassStandardCrashbreaker:convert(ClassStandardCrashbreaker, crashbreaker)
    end
end

local function reset()
    --[[
    Resets the crashbreaker and pickups for the next attempt
    ]]--
    resetCrashbreaker()
    resetPickups()
end

local function resetState()
    --[[
    Resets variables back to their initial state
    ]]--
    pickups = {coins = {}, multipliers = {}, explosives = {}}
    crashbreaker = nil
end

----------------------------------------------------------------------------------------------------
--Getters
----------------------------------------------------------------------------------------------------

local function getPickups()      return pickups      end
local function getCrashbreaker() return crashbreaker end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

M.createPickupInstances = createPickupInstances
M.checkCollectables = checkCollectables
M.update = update
M.reset = reset
M.setupCrashbreaker = setupCrashbreaker
M.getPickups = getPickups
M.getCrashbreaker = getCrashbreaker
M.resetState = resetState
return M