local M = {}

----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

--Classes
local ClassTimer = require("scenario/intersection/classes/classTimer")

--Handlers
local pickupHandler  = require("scenario/intersection/handlers/pickupHandler")
local vehicleHandler = require("scenario/intersection/handlers/vehicleHandler")
local errorHandler   = require("scenario/intersection/handlers/errorHandler")

----------------------------------------------------------------------------------------------------
--Variables
----------------------------------------------------------------------------------------------------

--The relevant timers
local timers = {
    preTimeout = ClassTimer:new(2, true ),
    timeout    = ClassTimer:new(5, false),
    crashtime  = ClassTimer:new(8, true )
}

--Used for error checking
local STATES = {"init", "ready", "running", "pre-timeout", "timeout", "finished"}

local scTimerState = "init"

local isInSlowMotion = false

----------------------------------------------------------------------------------------------------

local function setSlowMotion(slowmo)
    --[[
    Sets slow motion
    Parameters:
        slowmo - whether to start (true) or end slow motion (false)
    ]]--
    errorHandler.assertNil(slowmo)

    if slowmo then
        --Don't re-set bullettime
        if isInSlowMotion then
            return
        end

        bullettime.set(1/4)
        isInSlowMotion = true
    else
        --Don't re-set bullettime
        if not isInSlowMotion then
            return
        end

        bullettime.set(1)
        isInSlowMotion = false
    end
end

local function checkScenarioEnding()
    --[[
    Checks the conditions necessary to see if the scenario should start to timeout
    ]]--
    local crashbreaker = pickupHandler.getCrashbreaker()

    --Wait for player to use crashbreaker first
    if not crashbreaker or not crashbreaker:isLocked() and not crashbreaker:isUsed() then
        return false
    end

    local vehicles = vehicleHandler.getVehicles()

    --[[
    Condition:
      -(AI vehicles are crashed
           OR have been eliminated
           OR stationary)
       AND the player velocity is close to stationary
    ]]--
    for _, instance in pairs(vehicles.ais) do
        if not (instance:hasCrashed() or (instance:isStationary() or instance:isEliminated())) then
            return false
        end
    end

    local player = vehicles.player
    return player:getVelocity():length() <= 0.25
end

local function setScenarioState(state)
    --[[
    Sets the scenario timer state
    Parameters:
        state - the state to set to
    ]]--
    errorHandler.assertNil(state)
    errorHandler.assertValidElement(state, STATES, state.." is not a valid timer state")

    if state == "running" then
        --Reset the timer
        timers.preTimeout:reset()
    elseif state == "pre-timeout" then
        timers.preTimeout:setActive(true)
    elseif state == "timeout" then
        timers.timeout:setActive(true)
    end

    scTimerState = state
end

local function update(dt, dtScaled)
    --[[
    Updates the timer currently running
    Parameters:
        dt - the time since the last frame
        dtScaled - the time since the last frame, accounting for slowmotion
    Returns:
        <boolean> - whether the scenario should finish
    ]]--
    local running    = scTimerState == "running"
    local timeout    = scTimerState == "timeout"
    local preTimeout = scTimerState == "pre-timeout"

    local shouldTimeout = checkScenarioEnding()

    if running and shouldTimeout then
        --Start the timeout procedure
        setScenarioState("pre-timeout")
    elseif preTimeout and not shouldTimeout then
        --Stop and reset the timeout procedure
        setScenarioState("running")
    end

    if running or timeout or preTimeout then
        --Update all timers
        for _, timer in pairs(timers) do
            if timer:isActive() then
                local deltaTime = timer:hasSlowmotiom() and dtScaled or dt
                timer:update(deltaTime)
            end
        end

        if preTimeout and timers.preTimeout:hasFinished() then
            --Grace period is over so now start the actual countdown
            guihooks.trigger("i_countdown_end")
            setScenarioState("timeout")
        elseif timeout and timers.timeout:hasFinished() then
            setScenarioState("finished")
            return true
        elseif timers.crashtime:hasFinished() then
            --Reset back to initial state to prevent this always being true
            timers.crashtime:reset()
            setSlowMotion(false)
        end
    end

    return false
end

local function getScenarioState()
    return scTimerState
end

local function initCrashtime()
    --[[
    Sets crashtime to start
    ]]--
    --Reset
    timers.crashtime:reset()

    timers.crashtime:setActive(true)

    setSlowMotion(true)
end

local function checkForCrashtime(player, vehicle1, vehicle2)
    --[[
    Checks whether crashtime should be activated after a crash
    Parameters:
        player   - the player vehicle
        vehicle1 - the first vehicle involved in a crash
        vehicle2 - the second vehicle involved in a crash
    ]]--
    --[[False Condition 1: player hasn't crashed as isn't part of the collision
        False Condition 2: both collision vehicles have already crashed]]--
    local cond1 = not player:hasCrashed() and vehicle1:getId() ~= player:getId() and vehicle2:getId() ~= player:getId()
    local cond2 = vehicle1:hasCrashed() and vehicle2:hasCrashed()
    if cond1 or cond2 then
        return
    end

    initCrashtime()
end

local function reset()
    --[[
    Resets all timers and the scenario state back to its initial state
    ]]--
    for _, instance in pairs(timers) do
        instance:reset()
    end

    setScenarioState("ready")
    setSlowMotion(false)
end

local function resetState()
    --[[
    Resets variables back to their initial state
    ]]--
    timers = {
        preTimeout = ClassTimer:new(2, true ),
        timeout    = ClassTimer:new(5, false),
        crashtime  = ClassTimer:new(8, true )
    }

    scTimerState = "init"
    isInSlowMotion = false
end

M.setScenarioState = setScenarioState
M.update = update
M.getScenarioState = getScenarioState
M.initCrashtime = initCrashtime
M.checkForCrashtime = checkForCrashtime
M.reset = reset
M.resetState = resetState
return M