local M = {}

----------------------------------------------------------------------------------------------------
--Imports
----------------------------------------------------------------------------------------------------

--Classes
local ClassAi     = require("scenario/intersection/classes/vehicles/classAi")
local ClassProp   = require("scenario/intersection/classes/vehicles/classProp")
local ClassTarget = require("scenario/intersection/classes/vehicles/classTarget")
local ClassPlayer = require("scenario/intersection/classes/vehicles/classPlayer")
local ClassVector = require("scenario/intersection/classes/classVector")
--Miscellanous
local tableComp    = require("scenario/intersection/tableComprehension")
local errorHandler = require("scenario/intersection/handlers/errorHandler")

----------------------------------------------------------------------------------------------------
--Variables
----------------------------------------------------------------------------------------------------

local vehicles = {player = nil, ais = {}, props = {}}

----------------------------------------------------------------------------------------------------

local function findObjects(type)
    --[[
    Finds the objects of a particular type
    Parameters:
        type - the object type to find
    Returns:
        <table> - the objects found
    ]]--
    errorHandler.assertNil(type)
    errorHandler.assertValidElement(type, {"prop", "ai"}, type.." is not a valid object to find")

    local found = {}

    local i = 1
    local name = type..i
    local obj = scenetree.findObject(name)

    while obj do
        found[i] = {id = obj:getID(), name = name, position = obj:getPosition(), type = type}

        i = i + 1
        name = type..i
        obj = scenetree.findObject(name)
    end

    return found
end

local function findVehicles()
    --[[
    Finds the vehicles in the scenario
    Returns:
      vehicles - the list of vehicles, with their useful data
    ]]--
    local objsFound = findObjects("ai")

    --Add player in
    local playerVehicle = scenetree.findObject("scenario_player0")
    table.insert(objsFound, {
        id = playerVehicle:getID(),
        name = "scenario_player0",
        position = playerVehicle:getPosition(),
        type = "player"
    })

    --Add in props
    objsFound = tableComp.mergeAppend(objsFound, findObjects("prop"))

    return objsFound
end

local function createVehicleInstances(delays, target)
    --[[
    Creates instances of ClassVehicle from vehicle data
    Parameters:
        delay  - the delays, with values in seconds, each vehicle has (can be empty)
        target - the target vehicle (can be nil)
    ]]--
    errorHandler.assertNil(delays)

    local vehiclesFound = findVehicles()

    for _, data in ipairs(vehiclesFound) do
        local id   = data.id
        local pos  = data.position
        local name = data.name
        local type = data.type

        local vehData = {id, name, ClassVector:new(pos.x, pos.y, pos.z)}

        --Find appropriate instance
        local instance
        if type == "ai" then
            --AIs have the additional field 'delay'
            table.insert(vehData, delays[name] or 0)

            if name == target then
                instance = ClassTarget:new(unpack(vehData))
            else
                instance = ClassAi:new(unpack(vehData))
            end
        elseif type == "prop" then
            instance = ClassProp:new(unpack(vehData))
        elseif type == "player" then
            --Since field 'player' is not a table, skip adding it below and add it in manually
            vehicles.player = ClassPlayer:new(unpack(vehData))
            goto continue
        end

        --Set vehicles for collision tracking
        instance:trackDamage()

        vehicles[type.."s"][id] = instance

        ::continue::
    end
end

local function update(dt)
    --[[
    Updates the vehicles
    Parameters:
        dt - the time since the last frame
    ]]--
    for k, v in pairs(vehicles) do
        --Check for player as not a table
        if k == "player" then
            v:update(dt)
        else
            for _, instance in pairs(v) do
                instance:update(dt)
            end
        end
    end
end

local function start()
    --[[
    Handles the starting of the scenario for the vehicles, namely, setting them to active
    ]]--
    for _, instance in pairs(vehicles.ais) do
        instance:pend()
    end
end

local function reset()
    --[[
    Resets the vehicles for the next attempt
    ]]--
    for k, v in pairs(vehicles) do
        --Check for player as not a table
        if k == "player" then
            v:reset()
        else
            for _, instance in pairs(v) do
                instance:reset()
            end
        end
    end
end

local function getVehicles() return vehicles end

local function getVehicle(id)
    --[[
    Finds the vehicle by its id
    Parameters:
        id - the id of the vehicle to find
    Returns:
        <ClassVehicle> - the vehicle found
    ]]--
    errorHandler.assertNil(id)

    return vehicles.ais[id]
            or vehicles.props[id]
            or (vehicles.player:getId() == id and vehicles.player)
            or log("W", "vehicleHandler:getVehicle()", "No vehicle with id="..id.." exists")
end

local function getVehicleByName(name)
    --[[
    Finds the vehicle by its name
    Parameters:
        name - the name of the vehicle to find
    Returns:
        <ClassVehicle> - the vehicle found
    ]]--
    errorHandler.assertNil(name)

    for k, v in pairs(vehicles) do
        --Check for player as not a table
        if k == "player" then
            if v:getName() == name then
                return v
            end
        else
            for _, instance in pairs(v) do
                if instance:getName() == name then
                    return instance
                end
            end
        end
    end

    log("W", "vehicleHandler:getVehicleName()", name.." is not a valid vehicle")
end

local function resetState()
    --[[
    Resets variables back to their initial state
    ]]--
    vehicles = {player = nil, ais = {}, props = {}}
end

----------------------------------------------------------------------------------------------------
--Interface
----------------------------------------------------------------------------------------------------

M.createVehicleInstances = createVehicleInstances
M.update = update
M.start = start
M.reset = reset
M.getVehicles = getVehicles
M.getVehicle = getVehicle
M.getVehicleByName = getVehicleByName
M.resetState = resetState
return M
